module Main( main ) where

import System.Environment
import System.Console.GetOpt
import System.Exit

import Data.Either(lefts, rights)

import Control.Monad
import Data.List

import Latex
import Smt


{-- Data structures --}



{-- Structure: ExpInt --}
data ExpInt     = Number    Int
                | Variable  Int String
                | Array     Int String ExpInt
                | Function  String [ExpInt]
                | Neg       ExpInt
                | I_Plus    ExpInt ExpInt
                | I_Minus   ExpInt ExpInt
                | I_Times   ExpInt ExpInt
                | I_Div     ExpInt ExpInt
                | I_Mod     ExpInt ExpInt
    deriving Eq
	
{-- Instance of Show: ExpInt --}
instance Show (ExpInt) where
	show(Number   i)	= (show i)
	show(Variable i s)	= s++"_{"++(show i)++"}"
	show(Array	  i s e)= s++"_{"++(show i)++"}"++"["++(show e)++"]"
	show(Function s l)  = s++"("++(auxShowExpInt l)++")"
	show(Neg  e  ) 		=  "-" ++ show(e)
	show(I_Plus  e1 e2) = show(e1) ++ " + " ++ show(e2)
	show(I_Minus e1 e2) = show(e1) ++ " - " ++ show(e2)
	show(I_Times e1 e2) = show(e1) ++ " * " ++ show(e2)
	show(I_Div   e1 e2) = show(e1) ++ " div " ++ show(e2)
	show(I_Mod   e1 e2) = show(e1) ++ " mod " ++ show(e2)
auxShowExpInt [] 	= ""
auxShowExpInt [x] 	= (show x)
auxShowExpInt (h:t) = (show h)++", "++(auxShowExpInt t)

{-- Instância de reconhecimento: ExpInt --}
instance Read (ExpInt) where
	readsPrec _ = readsExpInt

readsExpInt s =	[]	
            ++[(I_Mod e1 e2, r) |
					("(",r1) 	        <- lex s,
					("I_Mod", r2) 	    <- lex r1,
					(e1, r3) 	        <- readsExpInt r2,
					(e2, r4) 	        <- readsExpInt r3,
					(")",r) 	        <- lex r4]
			++[(I_Div e1 e2, r) |
					("(",r1) 	        <- lex s,
					("I_Div", r2) 	    <- lex r1,
					(e1, r3) 	        <- readsExpInt r2,
					(e2, r4) 	        <- readsExpInt r3,
					(")",r) 	        <- lex r4]
			++[(I_Plus e1 e2, r) |
					("(",r1) 	        <- lex s,
					("I_Plus", r2) 	    <- lex r1,
					(e1, r3) 	        <- readsExpInt r2,
					(e2, r4) 	        <- readsExpInt r3,
					(")",r) 	        <- lex r4]
			++[(I_Minus e1 e2, r) |
					("(",r1) 	        <- lex s,
					("I_Minus", r2)     <- lex r1,
					(e1, r3) 	        <- readsExpInt r2,
					(e2, r4) 	        <- readsExpInt r3,
					(")",r) 	        <- lex r4]
			++[(I_Times e1 e2, r) |
					("(",r1) 	        <- lex s,
					("I_Times", r2)     <- lex r1,
					(e1, r3) 	        <- readsExpInt r2,
					(e2, r4) 	        <- readsExpInt r3,
					(")",r) 	        <- lex r4]
			++[(Neg e, r) |
					("(",r1) 	        <- lex s,
					("Neg", r2) 	    <- lex r1,
					( e, r3) 	        <- readsExpInt r2,
					(")",r) 	        <- lex r3]
			++[(Function s l, r) |
					("(",r1) 	        <- lex s,
					("Function", r2) 	<- lex r1,
					( s, r3) 	        <- lex r2,
					( l ,r4) 	        <- readList r3,
					(")",r) 	        <- lex r4]
			++[(Array 0 s e, r) |
					("(",r1) 	        <- lex s,
					("Array", r2) 	    <- lex r1,
					( s, r3) 	        <- lex r2,
					( e ,r4) 	        <- readsExpInt r3,
					(")",r) 	        <- lex r4]
			++[(Variable 0 s, r) |
					("(",r1) 	        <- lex s,
					("Variable", r2) 	<- lex r1,
					( s, r3) 	        <- lex r2,
					(")",r) 	        <- lex r3]
			++[(Number (read s::Int), r) |
					("(",r1) 	        <- lex s,
					("Number", r2) 	    <- lex r1,
					( s, r3) 	        <- lex r2,
					(")",r) 	        <- lex r3]
					








{-- Structure: ExpBool --}
data ExpBool    = B_Top
                | B_Bottom
                | B_Not  ExpBool
                | B_And  ExpBool ExpBool
                | B_Or   ExpBool ExpBool
                | B_Then ExpBool ExpBool
                | B_Eq   ExpInt  ExpInt
                | B_Neq  ExpInt  ExpInt
                | B_Le   ExpInt  ExpInt
                | B_Lt   ExpInt  ExpInt
                | B_Gt   ExpInt  ExpInt
                | B_Ge   ExpInt  ExpInt
    deriving Eq

{-- Instance of Show: ExpBool --}
instance Show ExpBool where
	show(B_Top)	       = " \\top "
	show(B_Bottom)	   = " \\bot "
	show(B_Not  b)	   = " \\lnot (" ++ (show b) ++ ")"
	show(B_And  b1 b2) = (show b1)++" \\land "++(show b2)
	show(B_Or   b1 b2) = (show b1)++" \\lor "++(show b2)
	show(B_Then b1 b2) = (show b1)++" \\rightarrow "++(show b2)
	show(B_Eq   e1 e2) = (show e1)++" \\equiv "++(show e2)
	show(B_Neq  e1 e2) = (show e1)++" {\\not\\equiv} "++(show e2)
	show(B_Lt   e1 e2) = (show e1)++" < "++(show e2)
	show(B_Le   e1 e2) = (show e1)++" \\le "++(show e2)
	show(B_Gt   e1 e2) = (show e1)++" > "++(show e2)
	show(B_Ge   e1 e2) = (show e1)++" \\ge "++(show e2)

{-- Instância de reconhecimento: ExpBool --}
instance Read (ExpBool) where
	readsPrec _ = readsExpBool
	
readsExpBool s = []	++[(B_Top, r)|
				("(",r1) 	    <- lex s,
				("B_Top",r2) 	<- lex r1,
				(")",r) 	    <- lex r2]
			++[(B_Bottom, r)|
				("(",r1) 	    <- lex s,
				("B_Bottom",r2) <- lex r1,
				(")",r) 	    <- lex r2]
			++[(B_Not b, r)|
				("(",r1) 	    <- lex s,
				("B_Not",r2) 	<- lex r1,
				( b ,r3) 	    <- readsExpBool r2,
				(")",r) 	    <- lex r3]
			++[(B_And b1 b2, r)|
				("(",r1) 	    <- lex s,
				("B_And",r2) 	<- lex r1,
				( b1,r3) 	    <- readsExpBool r2,
				( b2,r4) 	    <- readsExpBool r3,
				(")",r) 	    <- lex r4]
			++[(B_Or  b1 b2, r)|
				("(",r1) 	    <- lex s,
				("B_Or",r2) 	<- lex r1,
				( b1,r3) 	    <- readsExpBool r2,
				( b2,r4) 	    <- readsExpBool r3,
				(")",r) 	    <- lex r4]
			++[(B_Then  b1 b2, r)|
				("(",r1) 	    <- lex s,
				("B_Then",r2)   <- lex r1,
				( b1,r3) 	    <- readsExpBool r2,
				( b2,r4) 	    <- readsExpBool r3,
				(")",r) 	    <- lex r4]
			++[(B_Eq  e1 e2, r)|
				("(",r1)	    <- lex s,
				("B_Eq",r2) 	<- lex r1,
				( e1,r3) 	    <- readsExpInt r2,
				( e2,r4) 	    <- readsExpInt r3,
				(")",r) 	    <- lex r4]
			++[(B_Neq  e1 e2, r)|
				("(",r1)	    <- lex s,
				("B_Neq",r2) 	<- lex r1,
				( e1,r3) 	    <- readsExpInt r2,
				( e2,r4) 	    <- readsExpInt r3,
				(")",r) 	    <- lex r4]
			++[(B_Lt  e1 e2, r)|
				("(",r1) 	    <- lex s,
				("B_Lt",r2) 	<- lex r1,
				( e1,r3) 	    <- readsExpInt r2,
				( e2,r4) 	    <- readsExpInt r3,
				(")",r) 	    <- lex r4]
			++[(B_Le  e1 e2, r)|
				("(",r1) 	    <- lex s,
				("B_Le",r2) 	<- lex r1,
				( e1,r3) 	    <- readsExpInt r2,
				( e2,r4) 	    <- readsExpInt r3,
				(")",r) 	    <- lex r4]
			++[(B_Gt  e1 e2, r)|
				("(",r1) 	    <- lex s,
				("B_Gt",r2) 	<- lex r1,
				( e1,r3) 	    <- readsExpInt r2,
				( e2,r4) 	    <- readsExpInt r3,
				(")",r) 	    <- lex r4]
			++[(B_Ge  e1 e2, r)|
				("(",r1) 	    <- lex s,
				("B_Ge",r2) 	<- lex r1,
				( e1,r3) 	    <- readsExpInt r2,
				( e2,r4) 	    <- readsExpInt r3,
				(")",r) 	    <- lex r4]









{-- Structure: Command --}
data Command    = Skip
                | Assert ExpBool
                | Assign Int String  ExpInt
                | Comp               Command Command
                | If     ExpBool     Command Command
                | While  ExpBool     Command
                | Call   String
    deriving Eq

{-- Instance of Show: Command --}
instance Show Command where
	show c = pp_hs_command 0 c

pp_hs_command::Int->Command->String	
pp_hs_command n (Skip)			        = (pp_hs_tab n)++"skip";
pp_hs_command n (Assert	b)		        = (pp_hs_tab n)++"assert("++(show b)++")"
pp_hs_command n (Assign	i v e)		    = (pp_hs_tab n)++v++"_{"++(show i)++"}"++" = "++(show e);
pp_hs_command n (Comp        	c1 c2)	= (pp_hs_command n c1)++" ; \r\n"++(pp_hs_command n c2);
pp_hs_command n (If 	b	    c1 c2)	= (pp_hs_tab n)++"if("++(show b)++"){\r\n"++(pp_hs_command (n+1) c1)++"\r\n}else{\r\n"++(pp_hs_command (n+1) c2)++"\r\n}";
pp_hs_command n (While 	b 	    c)	    = (pp_hs_tab n)++"while("++(show b)++"){\r\n"++(pp_hs_command (n+1) c)++"\r\n"++(pp_hs_tab n)++"}";
pp_hs_command n (Call 	v)		        = (pp_hs_tab n)++"call "++( v)

latex_pp_command_w::Int->Command->String
latex_pp_command_w n c = ("")
    ++ nl("\\begin{spacing}{0.5}")
    ++ latex_pp_command n c
    ++ nl("\\end{spacing}")

latex_pp_command::Int->Command->String
latex_pp_command n (Skip)			    = (latex_pp_tab n)++" \\textbf{skip} "
latex_pp_command n (Call str)	        = (latex_pp_tab n)++" \\textbf{call}($"++(str)++"$) "
latex_pp_command n (Assert e)		    = (latex_pp_tab n)++" \\textbf{assert}($"++(show e)++"$) "
latex_pp_command n (Assign i v e)		= (latex_pp_tab n)++" $"++(v)++"_{"++(show i)++"} $ := $"++(show e)++"$ "
latex_pp_command n (Comp c1 c2)	        = (latex_pp_command n c1)++";~\\\\\n"++(latex_pp_command n c2)
latex_pp_command n (If b c1 c2)		    = (latex_pp_tab n)++" \\textbf{if}($"++(show b)++"$)\\{\\\\\n"
        ++(latex_pp_command (n+1) c1) ++ "\\\\\n"
        ++(latex_pp_tab n)++"\\}\\textbf{else}\\{\\\\\n"
        ++(latex_pp_command (n+1) c2) ++ "\\\\\n"
        ++(latex_pp_tab n)++"\\}\n "
latex_pp_command n (While b c)		= (latex_pp_tab n)++" \\textbf{while}($"++(show b)++"$) do\\{\\\\\n"
    ++(latex_pp_command (n+1) c) ++ "\\\\\n"
    ++(latex_pp_tab n)++"\\}\n "


{--
latex_pp_command n (Skip)			    = (latex_pp_tab n)++" \\textbf{skip} "
latex_pp_command n (Call str)	        = (latex_pp_tab n)++" \\textbf{call}($"++(str)++"$) "
latex_pp_command n (Assert e)		    = (latex_pp_tab n)++" \\textbf{assert}($"++(show e)++"$) "
latex_pp_command n (Assign i v e)		= (latex_pp_tab n)++" $ "++(v)++"_{"++(show i)++"} $  := $"++(show e)++"$ "
latex_pp_command n (Comp c1 c2)	        = (latex_pp_command n c1)++";~\\\\\n"++(latex_pp_command n c2)
latex_pp_command n (If b c1 c2)		    = (latex_pp_tab n)++" \\textbf{if}($"++(show b)++"$)\\{\\\\\n"
        ++(latex_pp_command (n+1) c1) ++ "\\\\\n"
        ++(latex_pp_tab n)++"\\}\\textbf{else}\\{\\\\\n"
        ++(latex_pp_command (n+1) c2) ++ "\\\\\n"
        ++(latex_pp_tab n)++"\\}\n "
latex_pp_command n (While b c)		= (latex_pp_tab n)++" \\textbf{while}($"++(show b)++"$) do\\{\\\\\n"
        ++(latex_pp_command (n+1) c) ++ "\\\\\n"
        ++(latex_pp_tab n)++"\\}\n "
--}

pp_hs_tab::Int->String
pp_hs_tab 0 = ""
pp_hs_tab n = "    "++(pp_hs_tab (n-1))

{-- Instância de reconhecimento: Command --}
instance Read (Command) where
	readsPrec _ = readsCommand
	
readsCommand s = []	
			++[(Skip, r) |	
            		("(",r1) 	    <- lex s,
            		("Skip", r2)    <- lex r1,
            		(")",r) 	    <- lex r2]
			++[(Assert b, r) |
					("(",r1) 	    <- lex s,
					("Assert", r2)  <- lex r1,
					( b, r3) 	    <- readsExpBool r2,
					(")",r) 	    <- lex r3]
			++[(Assign 0 s e2, r) | -- observe-se o valor(0), porque isto é um problema deste tipo de dados e não do parser, o parser desconhece isto.
					("(",r1) 	    <- lex s,
					("Assign", r2) 	<- lex r1,
					("(",r3) 	    <- lex r2,
					(s  ,r4) 	    <- lex r3,
					(")",r5) 	    <- lex r4,
					(e2, r6) 	    <- readsExpInt r5,
					(")",r) 	    <- lex r6]
			++[(Comp c1 c2, r) |
					("(",r1) 	    <- lex s,
					("Comp", r2)    <- lex r1,
					(c1, r3) 	    <- readsCommand r2,
					(c2, r4) 	    <- readsCommand r3,
					(")",r) 	    <- lex r4]
			++[(If b c1 c2, r) |
					("(",r1) 	    <- lex s,
					("If", r2) 	    <- lex r1,
					( b, r3) 	    <- readsExpBool r2,
					(c1, r4) 	    <- readsCommand r3,
					(c2, r5) 	    <- readsCommand r4,
					(")",r) 	    <- lex r5]
			++[(While b c, r) |
					("(",r1) 	    <- lex s,
					("While", r2) 	<- lex r1,
					( b, r3) 	    <- readsExpBool r2,
					( c, r4) 	    <- readsCommand r3,
					(")",r) 	    <- lex r4]
			++[(Call v, r) |
					("(",r1) 	    <- lex s,
					("Call", r2) 	<- lex r1,
					( v, r3) 	    <- lex r2,
					(")",r) 	    <- lex r3]
			

-- JSP extentions

data Procedure = Proc String Command
    deriving Eq

{-- Instance of Show: Procedure --}
instance Show Procedure where
    show(Proc str cmd) = "proc("++str++"){\n"++(show cmd)++"\n}\n"
--pp_hs_command n (Proc 	v c)		= (pp_hs_tab n)++"proc "++(show v)++" = {\r\n"++(pp_hs_command (n+1) c)++"\r\n}";

{-- Instância de reconhecimento: Procedure --}
instance Read (Procedure) where
	readsPrec _ = readsProcedure
	
readsProcedure s = []
			++[(Proc v c, r) |
					("(",r1) 	<- lex s,
					("Proc", r2) 	<- lex r1,
					( v, r3) 	<- lex r2,
					( c, r4) 	<- readsCommand r3,
					(")",r) 	<- lex r4]
--
--
-- Transformation 0
--
--
--
step0::[Procedure]->Command->Command
step0 l w@(Skip)         = (Skip)
step0 l w@(Assert b)     = (Assert b)
step0 l w@(Assign i v e) = (Assign i v e)
step0 l w@(Comp c1 c2)   = (Comp (step0 l c1) (step0 l c2))
step0 l w@(If b c1 c2)   = (If b (step0 l c1) (step0 l c2))
step0 l w@(While b c)    = (While  b          (step0 l c))
step0 l w@(Call  pName)
        |(length procedures == 1) = 
                (step0 l (head procedures))
        |(length procedures < 1) = 
                error("[ERROR]: procedure ["++pName++"] NOT found")
        |(length procedures > 1) = 
                error("[ERROR]: multiple decalarations of procedure ["++pName++"]")
    where procedures = (getProcedures pName l)

getProcedures::String->[Procedure]-> [Command]
getProcedures pid []    = []
getProcedures pid (h:t) = (hasProcedureId pid h) ++ (getProcedures pid t)

hasProcedureId pid (Proc name cmd)  | name==pid = [cmd]
                                    | otherwise = []


--
--
-- Transformation 1 - unWinding
--
--
--
step1::Int->Command->Command
step1 n w@(Skip)         = (Skip)
step1 n w@(Assert b)     = (Assert b)
step1 n w@(Assign i v e) = (Assign i v e)
step1 n w@(Comp c1 c2)   = (Comp (step1 n c1) (step1 n c2))
step1 n w@(If b c1 c2)   = (If b (step1 n c1) (step1 n c2))
step1 n w@(While b c)    = (uwd (n,n) w)
    where
        uwd (n,0) w@(While b cw) = (Assert (B_Bottom));
        uwd (n,m) w@(While b cw) = (If b (Comp (step1 n cw) (uwd (n,m-1) w)) Skip)
--
--
-- Transformation 2 - dynamic Single Assignment Form
--
--
--
update::(Memory,String)->Memory
update([],y) = [(y,1)] -- não é conhecido. Supondo que as variáveis já existem antes da execução
                       -- e supondo que esta rotina foi acionada por um Assert.
update(((x,i):t),y) | x==y = (x,i+1):t  -- é este
 					| True = (x,i):(update(t,y)) -- não é este

substituteExpInt::ExpInt->(String,Int)->ExpInt
substituteExpInt(Number   i)    (x,n) = (Number   i)
substituteExpInt(Variable i s)	(x,n)   |(s==x) = (Variable n x) -- puts n
                                        |(True) = (Variable i s) -- keep i  -- n!=i ?
substituteExpInt(Array	  i s e)(x,n) = (Array	  i s e)
substituteExpInt(Function s l)  (x,n) = (Function s l)
substituteExpInt(Neg  e)        (x,n) = (Neg     (substituteExpInt e  (x,n))                            )
substituteExpInt(I_Plus  e1 e2) (x,n) = (I_Plus  (substituteExpInt e1 (x,n)) (substituteExpInt e2 (x,n)))
substituteExpInt(I_Minus e1 e2) (x,n) = (I_Minus (substituteExpInt e1 (x,n)) (substituteExpInt e2 (x,n)))
substituteExpInt(I_Times e1 e2) (x,n) = (I_Times (substituteExpInt e1 (x,n)) (substituteExpInt e2 (x,n)))
substituteExpInt(I_Div   e1 e2) (x,n) = (I_Div   (substituteExpInt e1 (x,n)) (substituteExpInt e2 (x,n)))
substituteExpInt(I_Mod   e1 e2) (x,n) = (I_Mod   (substituteExpInt e1 (x,n)) (substituteExpInt e2 (x,n)))
substituteExpBool::ExpBool->(String,Int)->ExpBool
substituteExpBool(B_Top)	    (x,n) = (B_Top)
substituteExpBool(B_Bottom)	    (x,n) = (B_Bottom)
substituteExpBool(B_Not  b)	    (x,n) = (B_Not    (substituteExpBool b  (x,n)))
substituteExpBool(B_And  b1 b2) (x,n) = (B_And    (substituteExpBool b1 (x,n)) (substituteExpBool b2 (x,n)))
substituteExpBool(B_Or   b1 b2) (x,n) = (B_Or     (substituteExpBool b1 (x,n)) (substituteExpBool b2 (x,n)))
substituteExpBool(B_Then b1 b2) (x,n) = (B_Then   (substituteExpBool b1 (x,n)) (substituteExpBool b2 (x,n)))
substituteExpBool(B_Eq   e1 e2) (x,n) = (B_Eq     (substituteExpInt  e1 (x,n)) (substituteExpInt  e2 (x,n)))
substituteExpBool(B_Neq  e1 e2) (x,n) = (B_Neq    (substituteExpInt  e1 (x,n)) (substituteExpInt  e2 (x,n)))
substituteExpBool(B_Lt   e1 e2) (x,n) = (B_Lt     (substituteExpInt  e1 (x,n)) (substituteExpInt  e2 (x,n)))
substituteExpBool(B_Le   e1 e2) (x,n) = (B_Le     (substituteExpInt  e1 (x,n)) (substituteExpInt  e2 (x,n)))
substituteExpBool(B_Gt   e1 e2) (x,n) = (B_Gt     (substituteExpInt  e1 (x,n)) (substituteExpInt  e2 (x,n)))
substituteExpBool(B_Ge   e1 e2) (x,n) = (B_Ge     (substituteExpInt  e1 (x,n)) (substituteExpInt  e2 (x,n)))

--forAll l, f c
substituteExp f e l = foldl f e l
 
get::String->Memory->Int
get x [] = error "not in this memory"
get x ((y,i):t) |(x==y) = i
                |(True) = get x t



step2::Command->Command
step2 c = cmd
    where (tb,cmd) = dSAF [] c

type Memory = [(String,Int)]

dSAF::Memory->Command->(Memory,Command)
dSAF lIn w@(Skip)           = (lIn, (Skip))
dSAF lIn w@(Assert bIn)	    = (lIn,	(Assert bOut ))
    where
        bOut = (substituteExp substituteExpBool bIn lIn)    -- for all lIn, update values

dSAF lIn w@(Assign i x eIn)	= (lOut, (Assign n x eOut ))
    where
        eOut = (substituteExp substituteExpInt eIn lIn);    -- for all lIn, update values
        lOut = update(lIn,x);                               -- apply succ function
        n    = (get x lOut) {-- get value from update --}

dSAF lIn w@(Comp c1 c2)     = (lOut, (Comp cmd1 cmd2))
    where
        (lAux,cmd1) = (dSAF lIn c1);
        (lOut,cmd2) = (dSAF lAux c2)

dSAF lIn w@(If b c1 c2)		= (lIn, 
        (If (boolean)                                   -- updated boolean
            (foldl (\a b -> (Comp a b)) cmd1 lt)        -- join true
            (foldl (\a b -> (Comp a b)) cmd2 lf)  )     -- join false
    )
    where
        (boolean)     = (substituteExp substituteExpBool b lIn);
        (lTrue,cmd1)  = (dSAF lIn c1);
        (lFalse,cmd2) = (dSAF lIn c2);
        (lt,lf)       = unzip(compile lTrue lFalse);
        (lOut)        = (compose lTrue lFalse)          -- builds new table. Highest succ, win factor.

compose::Memory->Memory->Memory
compose m1 m2 = aux (m1++m2) []
    where
        aux []      acc = acc;
        aux (h:t)   acc = aux t (aux2 h acc);

        aux2 (x,i) []           = [(x,i)];-- nothing was found in this accumulator
        aux2 (x,i) ((y,n):t)    |(x==y) = (y,(max i n)+1):t
                                |(True) = (y,n):(aux2 (x,i) t)

compile::Memory->Memory->[(Command,Command)]
compile m1 m2 = map (aux m1 m2) (intersect (map (fst) m1) (map (fst) m2))
    where
        aux mTrue mFalse delta = let m = (max (get delta mTrue) (get delta mFalse)) in ((Assign (m+1) delta (Variable (get delta mTrue ) delta)), (Assign (m+1) delta (Variable (get delta mFalse) delta)) )


--
--
-- Transformation 3 - Conditional Normal Form
--
--
--
data CNF    = CNF_Comp 			CNF         CNF
            | CNF_IF 			ExpBool	  	Command
            
instance Show CNF where
	show(CNF_Comp m1 m2) = show(m1)++show(m2)
	show(CNF_IF   e  c ) = "if("++(show e)++")["++(show c)++"];\n"

latex_pp_CNF::CNF->String
latex_pp_CNF(CNF_Comp  m1 m2)  = (latex_pp_CNF m1)++";\\\\\n"++(latex_pp_CNF m2)
latex_pp_CNF(CNF_IF    b  c )  = " \\textbf{if}($"++(show b)++"$)\\{"
        ++(latex_pp_command 0 c)
        ++"\\} "











step3::Command -> ExpBool -> CNF
step3 p@(Skip)           phi = (CNF_IF (phi) (Skip))
step3 p@(Assert b)       phi = (CNF_IF (phi) (Assert b))
step3 p@(Assign i str e) phi = (CNF_IF (phi) (Assign i str e))
step3 p@(Comp ci cj)     phi = (CNF_Comp (step3 ci phi)  (step3 cj phi))
step3 p@(If b ct cf)     phi = (CNF_Comp 
        (step3 ct (B_And phi b))
        (step3 cf (B_And phi (B_Not b)))
    )





--
--
-- Transformation 4 - Conditional Normal Form
--
--
--
step4::CNF->([ExpBool],[ExpBool])
step4 c = ((lefts.skim) c, (rights.skim) c)
    where   
        skim(CNF_IF phi p@(Skip)          ) = [];
        skim(CNF_IF phi p@(Assert b)      ) = [(Right (B_Then phi b) )]; -- array and funcion normalization ignored
        skim(CNF_IF phi p@(Assign i str e)) = [(Left  (B_Then phi (B_Eq (Variable i str) e ) ) )];
        skim(CNF_Comp   ci cj             ) = (skim ci)++(skim cj)
        
setConjunction::[ExpBool]->ExpBool
setConjunction (h:t) = B_And h (setConjunction t)
setConjunction [] = B_Top


--
--
-- Main function
--
--
--
{-- is functions --}
isBound (Bound _) = True
isBound _ = False

isDsl (Dsl _) = True
isDsl _ = False





data Flag = Version | Dsl String | Bound Int deriving (Show,Eq)

options :: [OptDescr Flag]
options = [
    Option ['v'] ["version"] (NoArg Version)
                    "show version number",
    Option ['b'] ["bound"]   (ReqArg (\y-> Bound(read y::Int))  "34")
                    "BMC unWinding bound limit",
    Option ['f'] ["file"]    (ReqArg (\y-> Dsl y) "/Absolute/Path/file.dsl")
                    "input DSL file to read"
    ]

main = do{
    args <- getArgs;
    case (getOpt RequireOrder options args) of
        (listFlags, []     , []  ) -> parseOptions listFlags
        (_        , nonOpts, []  ) -> error $ "unrecognized arguments: "
                                            ++unwords nonOpts
        (_        , _      , msgs) -> error $ concat msgs
                                            ++ usageInfo "Usage: main [OPTION...]" options
}









parseOptions options = do{
    doVersion(options); -- print version if active                                      --[1]
    bound <- getBound options; -- get ONE bound if active                               --[2]
    dslPaths <- return(filter (\y -> isDsl y) options); -- filter dsl file paths        
    dslFileResults <- forM dslPaths (\y -> (parseDsl bound y)); -- apply bmc to files   --[3]
    forM dslFileResults pp_DslFileResults                                               --[4]
}



--[1]
doVersion::[Flag]->IO()
doVersion options   | length opt == 0   = return()
                    | otherwise         = putStrLn "Ferramenta Bounded Model Checking aplicado a software 2013" -- multiple version call
    where opt = filter (\y -> y == Version) options;





--[2]
getBound::[Flag]->IO(Int)
getBound l  |(length opt>1) = error "[error] please provide just one bound"
            |(length opt<1) = error "[error] please provide a bound"
            |(length opt==1) = return (extractBound opt)
    where   opt = filter (\y -> isBound y) l;
            extractBound [(Bound n)] = n



--[3]
data DslFileResult = DslFileResult String [ProcedureResult]
--    deriving Show
data ProcedureResult = ProcedureResult String Command Command Command CNF [ExpBool] [ExpBool]
--    deriving Show

{--
    1. from a string file path, 
    2. get contents,
    3. procedures is a list of read procedures,
    4. The file DslFileResult is a list of ProcedureResults
    
    Returns the result of applying bmc to ALL procedures in 1 file
--}
parseDsl::Int->Flag->IO(DslFileResult)
parseDsl bound (Dsl path) = do{
    text        <- readFile(path);
    procedures  <- return(map (\y->read y::Procedure) (lines text));       -- parse AST
    return(DslFileResult path (map (\y -> gbmc bound y procedures) procedures))
}
{--
Returns the result of applying bmc to 1 procedure
--}
gbmc::Int->Procedure->[Procedure]->ProcedureResult
gbmc bound (Proc name cmd) procedures = ProcedureResult name t0 t1 t2 t3 c p
    where
        t0 = (step0 procedures (Call name));
        t1 = (step1 bound t0);
        t2 = (step2 t1);
        t3 = (step3 t2 B_Top);
        (c,p) = (step4 t3);

--[4]
pp_DslFileResults::DslFileResult->IO()
pp_DslFileResults (DslFileResult absPath listProcedureResults) = do{
--    putStrLn ("i made ~> "++absPath++".tex :) im happy :) :)");
    list <- (forM listProcedureResults (mkLatexProcedurePage absPath)); -- esta cena vai criar o tex, percorrendo todos os smt
    writeFile (absPath++".tex") (pp_latex_2e_document [
        (pp_latex_2e_title 1 "-delete me-"),
        nl("\\tableofcontents"),
        concat( list ) -- para cada resultado de cada procedimento no ficheiro
    ]);
    putStrLn("Successfully generated tex file: "++absPath++".tex")
}


mkLatexProcedurePage::String->ProcedureResult->IO(String)
mkLatexProcedurePage absPath (ProcedureResult pid t0 t1 t2 cnf setC setP) = do{
    writeFile ((absPath)++".smt") (pp_smt_lib_v2_document(setC,setP));
    return(concat[
        nl("\\newpage"),
        nl("\\subsection{"++absPath++"}"),
        (pp_latex_2e_pair ("Procedimento("++pid++")",
            (latex_pp_command_w 0 t0)
        ) ("unWinding",
            (latex_pp_command_w 0 t1)
        ) ),
        nl("\\vfill"),
        (pp_latex_2e_pair ("Dynamic Single Assignment",
            (latex_pp_command_w 0 t2)
        ) ("Conditional Normal Form",
            (latex_pp_CNF cnf)
        ) ),
        nl("\\vfill"),
        (pp_latex_2e_pair ("Set $\\mathcal{C}$",
            (if((length setC)==0)then("$\\emptyset$")else(concat(map (\y->"$\\bullet$ $"++(show y)++"$~\\\\") setC)))
        ) ("Set $\\mathcal{P}$",
            (if((length setP)==0)then("$\\emptyset$")else(concat(map (\y->"$\\bullet$ $"++(show y)++"$~\\\\") setP)))
        ) ),
        (pp_latex_2e_block "SMT-LIB2"                   ("\\VerbatimInput[frame=lines,fontfamily=courier,fontshape=sl,fontsize=\\tiny]{"++(absPath)++".smt}") )
    ])
}



    
    

{-- SMT-LIB v2.0--}
pp_smt_lib_v2_document::([ExpBool],[ExpBool])->String
pp_smt_lib_v2_document (c,p) = ("")
                        ++ (pp_smt_lib_v2_headers)                                                  --[1] Smt.hs
						++ (pp_smt_lib_v2_options)                                                  --[2] Smt.hs
						++ (pp_smt_lib_v2_declarations_local 		(c,p))                          --[3]
--						++ (pp_smt_lib_v2_declaration_functions 	lf) -- no functions  to declare --[to do]
--						++ (pp_smt_lib_v2_declaration_predicates 	ld) -- no predicates to declare --[to do]
--						++ (pp_smt_lib_v2_declaration_axioms 		la) -- no axioms     to declare --[to do]
						++ (pp_smt_lib_v2_assertions 			    (c,p))                          --[7]
						++ (pp_smt_lib_v2_footers)                                                  --[8]


--[7]
pp_smt_lib_v2_assertions::([ExpBool],[ExpBool])->String
pp_smt_lib_v2_assertions (c,p)	= nl("")
--                ++ smt_lib_v2_comment ""
				++ smt_lib_v2_comment "Assertions:"
                ++ smt_lib_v2_comment "    set(C) -> set(P) === not_set(C) or set(P)"
                ++ smt_lib_v2_comment "    is there a not (set(C)->set(P))? not(not_set(C) or set(P)) === set(C) and not_set(P)"
                ++ nl("(assert (and ; set(C), the Codification of the Program")
                ++ nl(smtShowSetConjunction 1 (set_logical_conjunction c))
                ++ nl(" (not ; not_set(P) the Codification of the Conditions")
                ++ nl(smtShowSetConjunction 1 (set_logical_conjunction p))
                ++ nl(")))")


--[3]
pp_smt_lib_v2_declarations_local::([ExpBool],[ExpBool])->String
pp_smt_lib_v2_declarations_local (c,p) = nl("")
--            ++ smt_lib_v2_comment ""
			++ smt_lib_v2_comment "Declarations:"
--			++ smt_lib_v2_comment ""
			++ concat(nub ((concat c2)++(concat p2))) -- there may come to be some duplications in declarations, e.g., in(C)->in(P)
	where
	    c2 = map(expBool_get)(c);
	    p2 = map(expBool_get)(p)

expInt_get::ExpInt->[String]
expInt_get(Number    s  ) = []
expInt_get(Variable  i s) = [nl("(declare-const "++s++""++(show i)++" Int)")]
expInt_get(Array   i s e) = error "smt error"
expInt_get(Function  s l) = [nl("(declare-fun " ++ s ++ " (" ++(foldr (++)("")(map (\x->" Int ")(l) )) ++ ") Int)")]
expInt_get(Neg       e  ) = expInt_get(e)
expInt_get(I_Plus  e1 e2) = expInt_get(e1) ++ expInt_get(e2)
expInt_get(I_Minus e1 e2) = expInt_get(e1) ++ expInt_get(e2)
expInt_get(I_Times e1 e2) = expInt_get(e1) ++ expInt_get(e2)
expInt_get(I_Div   e1 e2) = expInt_get(e1) ++ expInt_get(e2)
expInt_get(I_Mod   e1 e2) = expInt_get(e1) ++ expInt_get(e2)

expBool_get::ExpBool->[String]
expBool_get(B_Top       ) = []
expBool_get(B_Bottom    ) = []
expBool_get(B_Not    b  ) = expBool_get(b)
expBool_get(B_And  b1 b2) = expBool_get(b1)++expBool_get(b2)
expBool_get(B_Or   b1 b2) = expBool_get(b1)++expBool_get(b2)
expBool_get(B_Then b1 b2) = expBool_get(b1)++expBool_get(b2)
expBool_get(B_Eq   e1 e2) = expInt_get(e1)++expInt_get(e2)
expBool_get(B_Neq  e1 e2) = expInt_get(e1)++expInt_get(e2)
expBool_get(B_Lt   e1 e2) = expInt_get(e1)++expInt_get(e2)
expBool_get(B_Le   e1 e2) = expInt_get(e1)++expInt_get(e2)
expBool_get(B_Gt   e1 e2) = expInt_get(e1)++expInt_get(e2)
expBool_get(B_Ge   e1 e2) = expInt_get(e1)++expInt_get(e2)










data SetConjunction = SC_Item ExpBool | SC_And ExpBool SetConjunction

smtShowSetConjunction::Int->SetConjunction->String
smtShowSetConjunction n (SC_Item b)       = (pp_hs_tab (n+1)) ++ (smtShowExpBool b) ++ "\n"
smtShowSetConjunction n (SC_And b set)    = ""
    ++ (pp_hs_tab n) ++ " (and " ++ "\n"
    ++ (pp_hs_tab (n+1)) ++ (smtShowExpBool b) ++ "\n"
    ++ (smtShowSetConjunction (n+0) set)
    ++ ")"
    
{-- Construct a show meta level of Conjunctions --}
set_logical_conjunction::[ExpBool]->SetConjunction
set_logical_conjunction [] = (SC_Item B_Top)
set_logical_conjunction [x] = (SC_Item x)
set_logical_conjunction (h:t) = (SC_And h (set_logical_conjunction t))


smtShowExpInt::ExpInt->String
smtShowExpInt (Number    i  )   = (show i) ++ " "
smtShowExpInt (Variable  i s)   = " " ++ s ++ "" ++ (show i) ++ " "
smtShowExpInt (Array	 i s e) = error "smt error"
smtShowExpInt (Function  s l)   = " (" ++ s ++ (foldr (++) ("") (map (smtShowExpInt ) l)) ++ ") " -- mal
smtShowExpInt (Neg       e  )   = " (- 0 " ++ (smtShowExpInt e)  ++ ") "
smtShowExpInt (I_Plus  e1 e2)   = " (+ "   ++ (smtShowExpInt e1) ++ (smtShowExpInt e2) ++ ") "
smtShowExpInt (I_Minus e1 e2)   = " (- "   ++ (smtShowExpInt e1) ++ (smtShowExpInt e2) ++ ") "
smtShowExpInt (I_Times e1 e2)   = " (* "   ++ (smtShowExpInt e1) ++ (smtShowExpInt e2) ++ ") "
smtShowExpInt (I_Div   e1 e2)   = " (div " ++ (smtShowExpInt e1) ++ (smtShowExpInt e2) ++ ") "
smtShowExpInt (I_Mod   e1 e2)   = " (mod " ++ (smtShowExpInt e1) ++ (smtShowExpInt e2) ++ ") "

smtShowExpBool::ExpBool->String
smtShowExpBool (B_Top       ) = " true "
smtShowExpBool (B_Bottom    ) = " false "
smtShowExpBool (B_Not    b  ) = " (not " 
                               ++ (smtShowExpBool b)
                               ++ ") "
smtShowExpBool (B_And  b1 b2) = " (and "
                               ++ (smtShowExpBool b1)
                               ++ (smtShowExpBool b2)
                               ++ ")"
smtShowExpBool (B_Or   b1 b2) = " (or "
                               ++ (smtShowExpBool b1)
                               ++ (smtShowExpBool b2)
                               ++ ")"
smtShowExpBool (B_Then b1 b2) = smtShowExpBool (B_Or (B_Not b1) b2)
smtShowExpBool (B_Eq   e1 e2) = " (= "
                               ++ (smtShowExpInt e1)
                               ++ (smtShowExpInt e2)
                               ++ ")"
smtShowExpBool (B_Neq  e1 e2) = smtShowExpBool (B_Not (B_Eq e1 e2))
smtShowExpBool (B_Lt   e1 e2) = " (< "
                               ++ (smtShowExpInt e1)
                               ++ (smtShowExpInt e2)
                               ++ ")"
smtShowExpBool (B_Le   e1 e2) = " (<= "
                               ++ (smtShowExpInt e1)
                               ++ (smtShowExpInt e2)
                               ++ ")"
smtShowExpBool (B_Gt   e1 e2) = " (> "
                               ++ (smtShowExpInt e1)
                               ++ (smtShowExpInt e2)
                               ++ ")"
smtShowExpBool (B_Ge   e1 e2) = " (>= "
                               ++ (smtShowExpInt e1)
                               ++ (smtShowExpInt e2)
                               ++ ")"