/*
 * Mestrado em Matemática e Computação
 * Programação Certificada
 *      - Ano lectivo de 2010/2011
 * Projecto de Programação Certificada
 * 
 * 
 * Lógica de Hoare para uma linguagem imperativa simples
 *      - Prova de correcção e implementação de um VCGen
 * 
 * 
 * Realizado por:
 *      - Isabel Maria Ventura Santos - pg17227@alunos.uminho.pt
 *      - José João Peixoto Pereira   - pg17312@alunos.uminho.pt
 * 
 * 
 * Professores Orientadores:
 *      - Jorge Sousa Pinto           - jsp@di.uminho.pt
 *      - Luís Filipe Ribeiro Pinto   - luis@math.uminho.pt
 */
 
 
/*
 *  Específicação YACC
 */
 
 
%{ /* @header */
        #include <stdio.h>          // input/output
        #include <string.h>         // string duplication
        char buffer[10*1024*1024];  // buffer de impressão para o pretty printer da AST
        
        /* Controlo de erros */
        int err_line = 0;           // linha do ficheiro onde se localiza o lookhead
        int err_column = 0;         // coluna do ficheiro onde se localiza o lookhaed
%}

%start root             /* produção raíz */

%union { char* v_str; }             // token reconhecido (representado por string)

/* Declarações: Simbolos não terminais */
%type   <v_str> root
//%type   <v_str> predicate function axiom ensure require 
%type   <v_str> procedure 
%type   <v_str> command if_else
//%type   <v_str> declaration declarations
%type   <v_str> exp_int exp_int_list arguments
%type   <v_str> exp_bool //exp_assertion

/* Declarações: Simbolos terminais */
//%token	<v_str> META_PREDICATE META_FUNCTION META_AXIOM META_ENSURES META_REQUIRES
%token	<v_str> SKIP ASSIGN COMPOSITION IF ELSE WHILE DO
//%token	<v_str> A_And A_Or A_Then A_Iff FORALL EXISTS
%token	<v_str> B_Top B_Bottom B_Not B_And B_Or B_Then B_Eq B_Neq B_Lt B_Le B_Gt B_Ge
%token	<v_str> VARIABLE NUMBER  I_Plus I_Minus I_Times I_Div I_Mod
%token	<v_str> '[' ']' '(' ')' '{' '}' ','
%token	<v_str> PROC
%token  <v_str> CALL ASSERT
%token	<v_str> ID

/*
root
        : declarations                                                          {;}
        ;

declarations
        : declarations declaration                                              {printf("%s\n",$2);}
        |                                                                       {;}
        ;

declaration
        : procedure                                                             {$$ = $1;}
        | predicate                                                             {$$ = $1;}
        | function                                                              {$$ = $1;}
        | axiom                                                                 {$$ = $1;}
        | require                                                               {$$ = $1;}
        | ensure                                                                {$$ = $1;}
        ;
        
/* meta informação *//*
predicate
        : META_PREDICATE VARIABLE NUMBER					{sprintf(buffer,"predicate(Predicate %s %s)"    , $2, $3     );      $$=strdup(buffer);}
        ;
function
        : META_FUNCTION VARIABLE NUMBER					    {sprintf(buffer,"function(Function %s %s)"      , $2, $3     );      $$=strdup(buffer);}
        ;
axiom
        : META_AXIOM    exp_assertion					    {sprintf(buffer,"axiom(Axiom %s)"               , $2         );      $$=strdup(buffer);}
        ;
require
        : META_REQUIRES VARIABLE ':' exp_assertion			{sprintf(buffer,"require(Require %s %s)"        , $2, $4     );      $$=strdup(buffer);}
        ;
ensure
        : META_ENSURES  VARIABLE ':' exp_assertion			{sprintf(buffer,"ensure(Ensure %s %s)"          , $2, $4     );      $$=strdup(buffer);}
        ;
        
/* procedimentos *//*
procedure
        : PROC VARIABLE '=' command                         {sprintf(buffer,"procedure(Procedure %s %s)"    , $2, $4     );      $$=strdup(buffer);}
        ;
*/
%%      /* beginning of rules section */






/* file root */
root
        : root procedure     {printf("%s\n",$2);}
        | procedure          {printf("%s\n",$1);}
        |                    {;}
        ;




/* procedures */
procedure
		: PROC '(' VARIABLE ')' '{' command '}'
            {sprintf(buffer,"(Proc %s %s)", $3, $6);$$=strdup(buffer);}
		;





/* commands expression grammar */
command
        : SKIP
            {sprintf(buffer,"(Skip)");                  $$=strdup(buffer);}

        | ASSERT '(' exp_bool ')'
            {sprintf(buffer,"(Assert %s)", $3);         $$=strdup(buffer);}

        | VARIABLE ASSIGN exp_int
            {sprintf(buffer,"(Assign(%s)%s)", $1, $3);  $$=strdup(buffer);}

        | command COMPOSITION command
            {sprintf(buffer,"(Comp%s%s)", $1, $3);      $$=strdup(buffer);}

        | IF    '(' exp_bool ')'    '{' command '}' if_else
            {sprintf(buffer,"(If%s%s%s)", $3, $6, $8 ); $$=strdup(buffer);}

        | WHILE '(' exp_bool ')' DO '{' command '}'
            {sprintf(buffer,"(While%s%s)", $3, $7);     $$=strdup(buffer);}

        | CALL VARIABLE
            {sprintf(buffer,"(Call %s)", $2);           $$=strdup(buffer);}
        ;
	
if_else
        : ELSE '{' command '}'                         {$$=$3;}
        |                                              {$$="(Skip)";}
        ;


/* integer expression grammar */
exp_int
        : NUMBER
            {sprintf(buffer,"(Number %s)",$1);          $$=strdup(buffer);}

        | VARIABLE
            {sprintf(buffer,"(Variable %s)",$1);        $$=strdup(buffer);}

        | VARIABLE '[' exp_int ']'
            {sprintf(buffer,"(Array %s %s)",$1,$3);     $$=strdup(buffer);}

        | VARIABLE arguments
            {sprintf(buffer,"(Function %s %s)",$1,$2);  $$=strdup(buffer);}

        | exp_int I_Plus  exp_int
            {sprintf(buffer,"(I_Plus%s%s)",$1,$3);      $$=strdup(buffer);}

        | exp_int I_Minus exp_int
            {sprintf(buffer,"(I_Minus%s%s)",$1,$3);     $$=strdup(buffer);}

        | exp_int I_Times exp_int
            {sprintf(buffer,"(I_Times%s%s)",$1,$3);     $$=strdup(buffer);}

        | exp_int I_Div   exp_int
            {sprintf(buffer,"(I_Div%s%s)",$1,$3);       $$=strdup(buffer);}

        | exp_int I_Mod   exp_int
            {sprintf(buffer,"(I_Mod%s%s)",$1,$3);       $$=strdup(buffer);}
        ;
	
arguments
        : '(' exp_int_list ')'
            {sprintf(buffer,"[%s]",$2);                 $$=strdup(buffer);}
        | '(' ')'
            {sprintf(buffer,"[]");                      $$=strdup(buffer);}
        ;
	
exp_int_list
        : exp_int ',' exp_int_list
            {sprintf(buffer,"%s,%s",$1,$3);             $$=strdup(buffer);}
        | exp_int
            {sprintf(buffer,"%s",$1);                   $$=strdup(buffer);}
        ;







/* boolean expression grammar */
exp_bool
        : B_Top
            {sprintf(buffer,"(B_Top)");                $$=strdup(buffer);}

        | B_Bottom
            {sprintf(buffer,"(B_Bottom)");             $$=strdup(buffer);}

        | B_Not exp_bool
            {sprintf(buffer,"(B_Not%s)"   , $2);       $$=strdup(buffer);}

        | exp_bool B_And  exp_bool
            {sprintf(buffer,"(B_And%s%s)" , $1,$3);    $$=strdup(buffer);}

        | exp_bool B_Or   exp_bool
            {sprintf(buffer,"(B_Or%s%s)"  , $1,$3);    $$=strdup(buffer);}

        | exp_bool B_Then exp_bool
            {sprintf(buffer,"(B_Then%s%s)", $1,$3);    $$=strdup(buffer);}

        | exp_int  B_Eq   exp_int
            {sprintf(buffer,"(B_Eq%s%s)"  , $1,$3);    $$=strdup(buffer);}

        | exp_int  B_Neq  exp_int
            {sprintf(buffer,"(B_Neq%s%s)" , $1,$3);    $$=strdup(buffer);}

        | exp_int  B_Lt   exp_int
            {sprintf(buffer,"(B_Lt%s%s)"  , $1,$3);    $$=strdup(buffer);}

        | exp_int  B_Le   exp_int
            {sprintf(buffer,"(B_Le%s%s)"  , $1,$3);    $$=strdup(buffer);}

        | exp_int  B_Gt   exp_int
            {sprintf(buffer,"(B_Gt%s%s)"  , $1,$3);    $$=strdup(buffer);}

        | exp_int  B_Ge   exp_int
            {sprintf(buffer,"(B_Ge%s%s)"  , $1,$3);    $$=strdup(buffer);}
        ;

/* asserções *//*
exp_assertion
        : exp_bool                                      {sprintf(buffer,"(A_Atom%s)", $1);              $$=strdup(buffer);}
        | exp_assertion A_And  exp_assertion            {sprintf(buffer,"(A_And%s%s)", $1, $3);         $$=strdup(buffer);}
        | exp_assertion A_Or   exp_assertion            {sprintf(buffer,"(A_Or%s%s)", $1, $3);          $$=strdup(buffer);}
        | exp_assertion A_Then exp_assertion            {sprintf(buffer,"(A_Then%s%s)", $1, $3);        $$=strdup(buffer);}
        | exp_assertion A_Iff  exp_assertion            {sprintf(buffer,"(A_Iff%s%s)", $1, $3);         $$=strdup(buffer);}
        | FORALL exp_int exp_assertion                  {sprintf(buffer,"(A_Iff%s%s)", $1, $3);         $$=strdup(buffer);}
        | EXISTS exp_int exp_assertion                  {sprintf(buffer,"(A_Iff%s%s)", $1, $3);         $$=strdup(buffer);}
        ;
*/
%%      /* beginning of function section */


/*
 * Função main
 *      Invoca o analisador da gramática.
 *      Produz a representação textual da AST no standard output
 */
int main(int argc, char**argv){
        yyparse();
        return 0;
}

/*
 * Função yyerror
 *      Modificada para apresentar a localzação do erro de syntax
 */ 
int yyerror(char *s){
        fprintf(stderr, "[error near: line(%d), column(%d):] %s\n", err_line, err_column, s);
        return 1;
}

/*
 * Função yywarp
 *      Apresenta o sucesso da computação
 */
int yywrap() {return(1);}